"""Ring buffer of players at the table."""
from whist.core.cards.hand import Hand
from whist.core.game.player_at_table import PlayerAtTable
from whist.core.scoring.team import Team


class PlayOrder:
    """
    Iterates over the players at the table.
    """

    def __init__(self, teams: list[Team]):
        self._size = len(teams) * len(teams[0].players)
        self._next_player = 0
        self._play_order: list[PlayerAtTable] = [None] * self._size
        for team_index, team in enumerate(teams):
            for player_index, player in enumerate(team.players):
                player_index = team_index + player_index * len(teams)
                self._play_order[player_index] = PlayerAtTable(player, Hand())

    def __eq__(self, other):
        if not isinstance(other, PlayOrder):
            return False
        return self._play_order == other._play_order

    def next_order(self) -> 'PlayOrder':
        """
        Create the order for the next hand.
        :rtype: PlayOrder
        """
        return PlayOrder._new_order(self)

    def next_player(self) -> PlayerAtTable:
        """
        Retrieves the next player who's turn it is.
        :rtype: PlayOrder
        """
        player: PlayerAtTable = self._play_order[self._next_player]
        self._next_player = (self._next_player + 1) % self._size
        return player

    # pylint: disable=protected-access
    @classmethod
    def _new_order(cls, old_order: 'PlayOrder'):
        instance = cls.__new__(cls)
        instance._play_order = old_order._play_order[1:] + old_order._play_order[:1]
        instance._size = len(instance._play_order)
        instance._next_player = 0
        return instance



"""Trick implementation"""
from whist.core.cards.card import Card, Suit
from whist.core.cards.stack import Stack
from whist.core.game.errors import NotPlayersTurnError, TrickDoneError
from whist.core.game.legal_checker import LegalChecker
from whist.core.game.player_at_table import PlayerAtTable
from whist.core.game.warnings import TrickNotDoneWarning, ServSuitFirstWarning


class Trick:
    """
    One round of where every player plays one card.
    """

    def __init__(self, play_order: list[PlayerAtTable], trump: Suit):
        self._play_order: list[PlayerAtTable] = play_order
        self._stack: Stack = Stack()
        self._trump = trump

    @property
    def done(self) -> bool:
        """
        Is the trick done.
        :return: True if trick is done else false.
        :rtype: bool
        """
        return len(self._stack) == len(self._play_order)

    @property
    def winner(self) -> PlayerAtTable:
        """
        Player how won the trick.
        :return: Player instance of the winner if the trick is done.
        Else raises TrickNotDoneWarning
        :rtype: Player
        """
        if not self.done:
            raise TrickNotDoneWarning()
        winner_card = self._stack.winner_card(self._trump)
        return self._play_order[self._stack.get_turn(winner_card)]

    def play_card(self, player: PlayerAtTable, card: Card) -> None:
        """
        One player plays one card. Which is put on top of the stack.
        :param player: Player who wants to play a card.
        :type player: Player
        :param card: Card which the player wants to play.
        :type card: Card
        :return: None if successful, else raises TrickDoneError if every player already played a
        card.
        Or NotPlayersTurnError if a player attempts to play card although it is not they turn.
        :rtype: None
        """
        turn = len(self._stack)
        if turn == len(self._play_order):
            raise TrickDoneError()
        if player != self._play_order[turn]:
            raise NotPlayersTurnError(player.player, self._play_order[turn].player)
        if not LegalChecker.check_legal(player.hand, self._stack.lead):
            raise ServSuitFirstWarning()

        self._stack.add(card)



"""Checks the legality of a move."""
from whist.core.cards.card import Card
from whist.core.cards.hand import Hand


# pylint: disable=too-few-public-methods
class LegalChecker:
    """
    Static legal checker.
    """

    @staticmethod
    def check_legal(hand: Hand, lead: Card) -> bool:
        """
        Checks if move is legal.
        :param hand: of the current player
        :type hand: Hand
        :param lead: the first played card
        :type lead: Card
        :return: True if legal else false
        :rtype: bool
        """
        first_card_played = lead is not None
        return not (first_card_played and hand.contain_suit(lead.suit))



"""
Match making tool.
"""
import random

from whist.core.session.userlist import UserList


# pylint: disable=too-few-public-methods
class RandomMatch:
    """
    Distributes the players randomly to teams.
    """
    _num_teams: int
    _team_size: int
    _users: UserList

    def __init__(self, num_teams: int, team_size: int, users: UserList):
        self._num_teams = num_teams
        self._team_size = team_size
        self._users = users

    def distribute(self) -> None:
        """
        For given parameter distributes the players to teams.
        :return: None
        :rtype: None
        """
        players = self._users.players
        teams: list = list(range(0, self._team_size)) * self._num_teams
        for player in players:
            team_id = random.choice(teams)
            self._users.change_team(player, team_id)
            teams.remove(team_id)



"""One Game of whist"""
from typing import Optional

from whist.core.game.hand import Hand
from whist.core.game.play_order import PlayOrder
from whist.core.scoring.score_card import ScoreCard
from whist.core.scoring.team import Team


class Game:
    """
    One Game of whist.
    """

    def __init__(self, teams: list[Team]):
        super().__init__()
        self.teams: list[Team] = teams
        self.win_score: int = 3
        self.score_card: ScoreCard = ScoreCard()
        self._current_hand: Optional[Hand] = None

    def next_hand(self) -> Hand:
        """
        Checks if the current hand is done and if so will return the next hand. If not it will
        return the current hand.
        :rtype: Hand
        """
        if self._current_hand is None:
            self._current_hand = Hand(PlayOrder(self.teams))
        elif self._current_hand.done:
            next_order = self._current_hand.next_play_order
            self._current_hand = Hand(next_order)
        return self._current_hand

    @property
    def done(self):
        """
        Check if game is done.
        :return: True if done else false
        :rtype: bool
        """
        return self.win_score <= self.score_card.max



"""Hand held by player."""
from typing import final, Iterable, Any, Iterator

from whist.core.cards.card import Suit, Card


@final
class Hand:
    """
    Hand of player during a game.
    """

    def __init__(self, *args: (tuple[Iterable[Card]], tuple[Card, ...])) -> None:
        """
        Constructor

        :param args: multiple cards or one card iterable
        """
        if len(args) == 1 and not isinstance(args[0], Card):
            self.__cards = {*args[0]}
        else:
            self.__cards = {*args}

    def __contains__(self, card: Card) -> bool:
        return card in self.__cards

    def __len__(self):
        return len(self.__cards)

    def __iter__(self) -> Iterator[Card]:
        return iter(self.__cards)

    def __str__(self) -> str:
        return str(self.__cards)

    def __repr__(self) -> str:
        return f'Hand(cards={self.__cards!r})'

    def __eq__(self, other: Any) -> bool:
        if self.__class__ is other.__class__:
            # pylint: disable=protected-access
            return self.__cards == other.__cards
        return NotImplemented

    def add(self, card: Card) -> None:
        """
        Add a card to this deck.

        :param card: card to add
        """
        if card in self.__cards:
            raise KeyError(f'{card} already in deck')
        self.__cards.add(card)

    def remove(self, card: Card) -> None:
        """
        Remove a card from this deck.

        :param card: card to remove
        """
        self.__cards.remove(card)

    @staticmethod
    def empty():
        """
        Creates a empty hand.
        :return: empty hand
        :rtype: Hand
        """
        return Hand()

    def contain_suit(self, suit: Suit) -> bool:
        """
        Checks if a card of a suit is still in the hand.
        :param suit: which should be checked
        :type suit: Suit
        :return: True if contains this suit else False
        :rtype: bool
        """
        if len(self.__cards) == 0:
            return False
        return any((card for card in self.__cards if card.suit == suit))



"""Hand of whist"""
import collections

from whist.core.cards.deck import Deck
from whist.core.game.play_order import PlayOrder
from whist.core.game.player_at_table import PlayerAtTable
from whist.core.game.trick import Trick
from whist.core.game.warnings import TrickNotDoneWarning


class Hand:
    """
    Hand of whist.
    """

    def __init__(self, play_order: PlayOrder):
        self._tricks: list[Trick] = []
        self._current_play_order: PlayOrder = play_order
        self._trump = None

    @property
    def done(self) -> bool:
        """
        Check if the hand is done.
        :return: True if the hand is done, else False
        :rtype: bool
        """
        return len(self._tricks) == 13 and self._tricks[-1]

    @property
    def next_play_order(self) -> PlayOrder:
        """
        Returns the next order of player for next hand.
        :rtype: PlayOrder
        """
        return self._current_play_order.next_order()

    def deal(self) -> Trick:
        """
        Deals the hand and starts the first trick.
        :return: the first trick
        :rtype: Trick
        """
        deck = Deck.full()
        while len(deck) > 0:
            for player in self._current_play_order:
                card = deck.pop_random()
                if len(deck) == 1:
                    self._trump = card.suit
                player.hand.add(card)
        if self._trump is None:
            raise ValueError
        first_trick = Trick(self._current_play_order, self._trump)
        self._tricks.append(first_trick)
        return first_trick

    def next_trick(self) -> Trick:
        """
        Starts the next trick.
        :return: the next trick
        :rtype: Trick
        """
        if not self._tricks[-1].done:
            raise TrickNotDoneWarning()
        self._winner_plays_first_card()
        next_trick = Trick(self._current_play_order, trump=self._trump)
        self._tricks.append(next_trick)
        return next_trick

    def _winner_plays_first_card(self):
        winner: PlayerAtTable = self._tricks[-1].winner
        deque = collections.deque(self._current_play_order)
        rotation: int = self._current_play_order.index(winner)
        deque.rotate(rotation)
        self._current_play_order = list(deque)



"""
Handles users joining and leaving a table.
"""
from typing import Optional, Dict

from pydantic import BaseModel

from whist.core.user.player import Player
from whist.core.user.status import Status


class UserList(BaseModel):
    """
    User handler for tables.
    """
    users: Dict[Player, Status] = {}

    def __len__(self):
        return len(self.users)

    @property
    def players(self) -> list[Player]:
        """
        Returns all players at the table.
        :return: players of the table
        :rtype: list[Player]
        """
        return list(self.users.keys())

    @property
    def ready(self) -> bool:
        """
        Returns if all players are ready.
        :return: Ready or not
        :rtype: boolean
        """
        player_status: Status
        for player_status in self.users.values():
            if not player_status.ready:
                return False
        return True

    def team(self, player: Player) -> Optional[int]:
        """
        Gets the id of the team for a player.
        :param player: for which the id should be retrieved
        :type player: Player
        :return: Integer if player joined a team or None if not.
        :rtype: int
        """
        status: Status = self.users.get(player)
        return status.team

    def team_size(self, team: int) -> int:
        """
        Gets the size of the team.
        :param team: ID of the team
        :type team: int
        :return: Amount of members
        :rtype: int
        """
        return len([status for status in self.users.values() if status.team == team])

    def is_joined(self, player: Player) -> bool:
        """
        Checks if the player is already at the table.
        :param player: to check
        :type player: Player
        :return: True if is member else false
        :rtype: bool
        """
        return player in self.users.keys()

    def append(self, player: Player):
        """
        Adds a player to the list.
        :param player: player to join
        :type player: Player
        :return: None
        :rtype: None
        """
        if not self.is_joined(player):
            self.users.update({player: Status()})

    def remove(self, player: Player):
        """
        Removes the player from the list.
        :param player: player to leave
        :type player: Player
        :return: None
        :rtype: None
        """
        if self.is_joined(player):
            self.users.pop(player)

    def change_team(self, player: Player, team: int) -> None:
        """
        Player changes teams.
        :param player: to change teams
        :type player: Player
        :param team: id of the new team
        :type team: int
        :return: None
        :rtype: None
        """
        status: Status = self.users.get(player)
        status.team = team

    def player_ready(self, player: Player):
        """
        Player says they is ready.
        :param player: player who is ready
        :type player: Player
        :return: None
        :rtype: None
        """
        status: Status = self.users.get(player)
        status.ready = True

    def player_unready(self, player: Player):
        """
        Player says they is not ready.
        :param player: player who is not ready
        :type player: Player
        :return: None
        :rtype: None
        """
        status: Status = self.users.get(player)
        status.ready = False



"""Errors regarding table logic."""


class TableFullError(Exception):
    """
    Thrown if the table is already full.
    """


class TeamFullError(Exception):
    """
    Thrown if the team is already full.
    """



"""Player instance during the game phase."""
from whist.core.cards.hand import Hand
from whist.core.user.player import Player


class PlayerAtTable:
    """
    Wraps the current hand and player instance.
    """

    def __init__(self, player: Player, hand: Hand):
        self._player = player
        self._hand = hand

    def __eq__(self, other):
        if not isinstance(other, PlayerAtTable):
            return False
        return self._player == other._player

    def __repr__(self):
        return f'PlayerAtTable: {self._player}'

    @property
    def player(self):
        """
        Getter player instance.
        :return: player instance
        :rtype: Player
        """

        return self._player

    @property
    def hand(self) -> Hand:
        """
        Hand of the player
        :return: hand of the player
        :rtype: Hand
        """
        return self._hand



"""Collection of cards"""
from typing import Optional

from whist.core.cards.card import Card, Suit


class Stack:
    """An ordered collection of cards"""

    def __init__(self):
        self.__cards: list[Card] = []

    def __len__(self):
        return len(self.__cards)

    @property
    def lead(self) -> Optional[Card]:
        """
        Returns the first card played.
        :return: The first card played if it exists. Else none.
        :rtype: Card or None
        """
        if len(self.__cards) == 0:
            return None
        return self.__cards[0]

    def add(self, card: Card) -> None:
        """
        Put card on top of stack
        :param card: a Card placed on top of the stack
        """
        if card in self.__cards:
            raise KeyError(f'{card} already in deck')
        self.__cards.append(card)

    def winner_card(self, trump: Suit) -> Card:
        """
        Returns the highest trump card or the highest card of the suit played first.
        :param trump: suit of trump
        :type trump: Suit
        :return: the winning card
        :rtype: Card
        """
        winner_suit_cards: list[Card] = self._card_of_suit(trump)
        if len(winner_suit_cards) == 0:
            winner_suit_cards: list[Card] = self._card_of_suit(self.__cards[0].suit)
        return max(winner_suit_cards, key=lambda x: x.rank)

    def get_turn(self, card: Card) -> int:
        """
        Gets the turn of a card played.
        :param card: for which the turn number shall be found
        :type card: Card
        :return: the index of the card in the stack. 0 is the first the card played
        :rtype: int
        """
        if card not in self.__cards:
            raise KeyError(f'{card} is not in stack.')
        return self.__cards.index(card)

    def _card_of_suit(self, trump):
        return [card for card in self.__cards if card.suit == trump]
